import numpy as np
import pandas as pd
from scipy.stats import ks_2samp
from sklearn.decomposition import PCA


def calculate_shape_stats(
    pca: PCA, data: pd.DataFrame, ref_data: pd.DataFrame, components: int
) -> pd.DataFrame:
    all_stats = []

    columns = ref_data.filter(like="shcoeffs").columns
    data_transform = pca.transform(data[columns].values)
    ref_data_transform = pca.transform(ref_data[columns].values)

    for component in range(components):
        ks_stats = get_ks_statistic(data_transform[:, component], ref_data_transform[:, component])
        ks_stats.update({"FEATURE": f"PC_{component + 1}", "TICK": np.nan})
        all_stats.append(ks_stats)

    for tick, tick_data in data.groupby("TICK"):
        tick_data_transform = pca.transform(tick_data[columns].values)

        for component in range(components):
            tick_ks_stats = get_ks_statistic(
                tick_data_transform[:, component], ref_data_transform[:, component]
            )
            tick_ks_stats.update(
                {
                    "FEATURE": f"PC_{component + 1}",
                    "TICK": tick,
                }
            )
            all_stats.append(tick_ks_stats)

    all_stats_df = pd.DataFrame(all_stats)

    return all_stats_df


def get_ks_statistic(population_a: np.ndarray, population_b: np.ndarray) -> dict:
    ksresult = ks_2samp(population_a, population_b, mode="asymp")

    return {
        "KS_STATISTIC": ksresult.statistic,
        "KS_PVALUE": ksresult.pvalue,
    }



import tempfile

import numpy as np
import pandas as pd
import trimesh
from aicsshparam import shtools
from sklearn.decomposition import PCA
from vtkmodules.vtkCommonDataModel import vtkPolyData
from vtkmodules.vtkIOPLY import vtkPLYWriter


def extract_shape_modes(
    pca: PCA, data: pd.DataFrame, components: int, regions: list[str], order: int, delta: float
) -> dict:
    features = data.filter(like="shcoeffs").columns.values
    data_transform = pca.transform(data[features].values)
    means = data_transform.mean(axis=0)
    stds = data_transform.std(axis=0, ddof=1)

    shape_svgs = {}

    for component in range(components):
        point_vector = np.zeros((components))
        component_shape_modes = {}

        for point in np.arange(-2, 2.5, delta):
            point_vector[component] = point
            vector = means + np.multiply(stds, point_vector)
            point_shape_modes = {}

            for region in regions:
                shape_mode_slices = extract_shape_mode_slices(pca, vector, features, order, region)
                point_shape_modes[region] = shape_mode_slices

            component_shape_modes[point] = point_shape_modes

        shape_svgs[component + 1] = component_shape_modes

    return shape_svgs


def extract_shape_mode_slices(
    pca: PCA, vector: np.ndarray, feature_names: list[str], order: int, region: str
) -> dict:
    prefix = ""
    suffix = f".{region}" if region != "DEFAULT" else ""
    mesh = construct_mesh_from_points(pca, vector, feature_names, order, prefix, suffix)
    mesh = convert_vtk_to_trimesh(mesh)
    slices = get_mesh_slices(mesh)
    return slices


def construct_mesh_from_points(
    pca: PCA,
    points: np.ndarray,
    feature_names: list[str],
    order: int,
    prefix: str = "",
    suffix: str = "",
) -> vtkPolyData:
    """Constructs mesh given PCA transformation points."""
    coeffs = pd.Series(pca.inverse_transform(points), index=feature_names)
    coeffs_map = np.zeros((2, order + 1, order + 1), dtype=np.float32)

    for l in range(order + 1):
        for m in range(order + 1):
            coeffs_map[0, l, m] = coeffs[f"{prefix}shcoeffs_L{l}M{m}C{suffix}"]
            coeffs_map[1, l, m] = coeffs[f"{prefix}shcoeffs_L{l}M{m}S{suffix}"]

    mesh, _ = shtools.get_reconstruction_from_coeffs(coeffs_map)
    return mesh


def convert_vtk_to_trimesh(mesh: vtkPolyData) -> trimesh.Trimesh:
    with tempfile.NamedTemporaryFile() as temp:
        writer = vtkPLYWriter()
        writer.SetInputData(mesh)
        writer.SetFileTypeToASCII()
        writer.SetFileName(f"{temp.name}.ply")
        _ = writer.Write()
        mesh = trimesh.load(f"{temp.name}.ply")

    return mesh


def get_mesh_slices(mesh: trimesh.Trimesh) -> dict:
    return {
        "side_1": get_mesh_slice(mesh, (0, 1, 0)),
        "side_2": get_mesh_slice(mesh, (1, 0, 0)),
        "top": get_mesh_slice(mesh, (0, 0, 1)),
    }


def get_mesh_slice(mesh: trimesh.Trimesh, normal: tuple[int, int, int]) -> str:
    """Get svg slice of mesh along plane for given normal."""
    mesh_slice = mesh.section_multiplane(mesh.centroid, normal, [0])
    svg = trimesh.path.exchange.svg_io.export_svg(mesh_slice[0])
    return svg



import numpy as np
import pandas as pd

from abm_shape_collection.calculate_shape_stats import get_ks_statistic


def calculate_size_stats(
    data: pd.DataFrame, ref_data: pd.DataFrame, regions: list[str]
) -> pd.DataFrame:
    all_stats = []

    for region in regions:
        for feature in ["volume", "height"]:
            column_name = f"{feature}.{region}" if region != "DEFAULT" else feature
            ref_values = ref_data[column_name].values
            values = data[column_name].values

            ks_stats = get_ks_statistic(values, ref_values)
            ks_stats.update({"FEATURE": column_name, "TICK": np.nan})
            all_stats.append(ks_stats)

            for tick, tick_data in data.groupby("TICK"):
                tick_values = tick_data[column_name].values

                tick_ks_stats = get_ks_statistic(tick_values, ref_values)
                tick_ks_stats.update(
                    {
                        "FEATURE": column_name,
                        "TICK": tick,
                    }
                )
                all_stats.append(tick_ks_stats)

    all_stats_df = pd.DataFrame(all_stats)

    return all_stats_df



import importlib
import os
import sys

from prefect import task

for module_file in os.listdir(os.path.dirname(__file__)):
    if "__" in module_file or not module_file.endswith(".py"):
        continue

    module_name = module_file.replace(".py", "")

    module = importlib.import_module(f".{module_name}", package=__name__)
    setattr(sys.modules[__name__], module_name, task(getattr(module, module_name)))
